/* 
               Central DCC Bluetooth
               Joaquim Herdy Menezes 
                   < Joahermen >
                      2019C28
                      
            Original DCC++ BASE STATION
               Autor Gregg E. Berman
              CopyRight (c) 2013-2016 
              
         *********************************
         *   Criador Genial do Arduino   *
         *        ===============        *
         *     <<  Massimo Banzi  >>     *
         *        ===============        *
         *********************************

      Modulos:
                Arduino Uno R3
                MotorShield L298 R3
                Bluetooth HC-06 ou HC-05 
    Observacsoes:
    Observacsoes:
    Em hipotese alguma conectar o modulo Bluetooth nos pinos D0 e D1 do Arduino 
        enquanto ele estiver conectado via cabo USB para os testes
    isolar o pino Vin do MotorShield L298R da ligacsao do Vin do Arduino 
    Fazer as conexoes dos pinos D5, D10, D12, D13 no interruptor Seletor conforme o esquema 
    ou sem interruptor Seletor jumpar o pino D5 com o pino D13 e o pino D10 com o pino D12
*/

#include "DCCpp_Uno.h"
#include "PacketRegister.h"
#include "CurrentMonitor.h"
#include "Sensor.h"
#include "SerialCommand.h"
#include "Accessories.h"
#include "EEStore.h"
#include "Config.h"
#include "Comm.h"
#include <SoftwareSerial.h>

//                                      A T E N S A O: 
// O Bluetooth usando os pinos D0 e D1 do Arduino, obrigatoriamente, ele nao pode estar conectado pelo Cabo via USB 
SoftwareSerial Bluetooth(0,1);                                // Bluetooth Pino Rx <Resistor> no Pino D1 do Arduino e Pino Tx no Pino D0 do Arduino

void showConfiguration();

#if COMM_TYPE == 1
  byte mac[] =  MAC_ADDRESS;                                  // Create MAC address (to be used for DHCP when initializing server)
  EthernetServer INTERFACE(ETHERNET_PORT);                    // Create and instance of an EnternetServer
#endif

volatile RegisterList mainRegs(MAX_MAIN_REGISTERS);           // create list of registers for MAX_MAIN_REGISTER Main Track Packets
volatile RegisterList progRegs(2);                            // create a shorter list of only two registers for Program Track Packets

CurrentMonitor mainMonitor(CURRENT_MONITOR_PIN_MAIN,"<p2>");  // create monitor for current on Main Track
CurrentMonitor progMonitor(CURRENT_MONITOR_PIN_PROG,"<p3>");  // create monitor for current on Program Track

void loop(){
  SerialCommand::process();                                   // check for, and process, and new serial commands
  if(CurrentMonitor::checkTime()){                            // if sufficient time has elapsed since last update, check current draw on Main and Program Tracks 
    mainMonitor.check();
    progMonitor.check();
  }
  Sensor::check();    // check sensors for activate/de-activate
} 

void setup(){  
  Serial.begin(9600);                                           // Frequencia interface serial
//  Serial.begin(57600);                                           // Frequencia interface serial - Funciona via USB
//  Serial.begin(115200);                                          // Nessa frequencia Monitor apresenta caracteres estranhos
  Bluetooth.begin(9600);                                        // Seta Frequencia para o Bluetooth
//  Bluetooth.begin(57600);                                        // Nao funcionou para o Bluetooth APP sem configuracsao de frequencias
//  Bluetooth.begin(115200);                                       // Arduino UNO nao alcansa essa frequencia
  Serial.flush();
  #ifdef SDCARD_CS
  pinMode(SDCARD_CS,OUTPUT);
  digitalWrite(SDCARD_CS,HIGH);                                 // Deselect the SD card
  #endif
  EEStore::init();                                              // initialize and load Turnout and Sensor definitions stored in EEPROM
  pinMode(A3,INPUT);                                            // if pin A3 is grounded upon start-up, print system configuration and halt
  digitalWrite(A3,HIGH);
  if(!digitalRead(A3))
    showConfiguration();
  Serial.print(" = Central DCC RealJoia Bluetooth =\n < Arduino ");         // Print Status to Serial Line regardless of COMM_TYPE setting so use can open Serial Monitor and check configurtion 
  Serial.print(ARDUINO_TYPE);
  Serial.println(" >");
  Serial.print(" > ");
  Serial.print(MOTOR_SHIELD_NAME);
  Serial.print(": V-");
  Serial.print(VERSION);
  Serial.println(" < ");
  Serial.print(" | ");
  Serial.print(__DATE__);
  Serial.print(" ");
  Serial.print(__TIME__);
  Serial.println(" | ");
  #if COMM_TYPE == 1
    #ifdef IP_ADDRESS
      Ethernet.begin(mac,IP_ADDRESS);                           // Start networking using STATIC IP Address
    #else
      Ethernet.begin(mac);                                      // Start networking using DHCP to get an IP Address
    #endif
    INTERFACE.begin();
  #endif
  SerialCommand::init(&mainRegs, &progRegs, &mainMonitor);      // create structure to read and parse commands from serial line
    Serial.print("<N ");
    Serial.print(COMM_TYPE);
  #if COMM_TYPE == 0
    Serial.println(" Serial>");
  #elif COMM_TYPE == 1
    Serial.print(" ");
    Serial.print(Ethernet.localIP());
    Serial.println(" >");
  #endif
  #define DCC_ZERO_BIT_TOTAL_DURATION_TIMER1 3199
  #define DCC_ZERO_BIT_PULSE_DURATION_TIMER1 1599
  #define DCC_ONE_BIT_TOTAL_DURATION_TIMER1 1855
  #define DCC_ONE_BIT_PULSE_DURATION_TIMER1 927
  pinMode(DIRECTION_MOTOR_CHANNEL_PIN_A,INPUT);                 // ensure this pin is not active! Direction will be controlled by DCC SIGNAL instead (below)
  digitalWrite(DIRECTION_MOTOR_CHANNEL_PIN_A,LOW);
  pinMode(DCC_SIGNAL_PIN_MAIN, OUTPUT);       // THIS ARDUINO OUPUT PIN MUST BE PHYSICALLY CONNECTED TO THE PIN FOR DIRECTION-A OF MOTOR CHANNEL-A
  bitSet(TCCR1A,WGM10);                                         // set Timer 1 to FAST PWM, with TOP=OCR1A
  bitSet(TCCR1A,WGM11);
  bitSet(TCCR1B,WGM12);
  bitSet(TCCR1B,WGM13);
  bitSet(TCCR1A,COM1B1);                      // set Timer 1, OC1B (pin 10/UNO, pin 12/MEGA) to inverting toggle (actual direction is arbitrary)
  bitSet(TCCR1A,COM1B0);
  bitClear(TCCR1B,CS12);                                        // set Timer 1 prescale=1
  bitClear(TCCR1B,CS11);
  bitSet(TCCR1B,CS10);
  OCR1A=DCC_ONE_BIT_TOTAL_DURATION_TIMER1;
  OCR1B=DCC_ONE_BIT_PULSE_DURATION_TIMER1;
  pinMode(SIGNAL_ENABLE_PIN_MAIN,OUTPUT);                       // master enable for motor channel A
  mainRegs.loadPacket(1,RegisterList::idlePacket,2,0);          // load idle packet into register 1    
  bitSet(TIMSK1,OCIE1B);                                        // enable interrupt vector for Timer 1 Output Compare B Match (OCR1B)    
#ifdef ARDUINO_AVR_UNO                                          // Configuration for UNO
  #define DCC_ZERO_BIT_TOTAL_DURATION_TIMER0 49
  #define DCC_ZERO_BIT_PULSE_DURATION_TIMER0 24
  #define DCC_ONE_BIT_TOTAL_DURATION_TIMER0 28
  #define DCC_ONE_BIT_PULSE_DURATION_TIMER0 14
  pinMode(DIRECTION_MOTOR_CHANNEL_PIN_B,INPUT);      // ensure this pin is not active! Direction will be controlled by DCC SIGNAL instead (below)
  digitalWrite(DIRECTION_MOTOR_CHANNEL_PIN_B,LOW);
  pinMode(DCC_SIGNAL_PIN_PROG,OUTPUT);      // THIS ARDUINO OUTPUT PIN MUST BE PHYSICALLY CONNECTED TO THE PIN FOR DIRECTION-B OF MOTOR CHANNEL-B
  bitSet(TCCR0A,WGM00);                                         // set Timer 0 to FAST PWM, with TOP=OCR0A
  bitSet(TCCR0A,WGM01);
  bitSet(TCCR0B,WGM02);
  bitSet(TCCR0A,COM0B1);                                        // set Timer 0, OC0B (pin 5) to inverting toggle (actual direction is arbitrary)
  bitSet(TCCR0A,COM0B0);
  bitClear(TCCR0B,CS02);                                        // set Timer 0 prescale=64
  bitSet(TCCR0B,CS01);
  bitSet(TCCR0B,CS00);
  OCR0A=DCC_ONE_BIT_TOTAL_DURATION_TIMER0;
  OCR0B=DCC_ONE_BIT_PULSE_DURATION_TIMER0;
  pinMode(SIGNAL_ENABLE_PIN_PROG,OUTPUT);   // master enable for motor channel B
  progRegs.loadPacket(1,RegisterList::idlePacket,2,0);    // load idle packet into register 1    
  bitSet(TIMSK0,OCIE0B);    // enable interrupt vector for Timer 0 Output Compare B Match (OCR0B)
#else                                                     // Configuration for MEGA
  #define DCC_ZERO_BIT_TOTAL_DURATION_TIMER3 3199
  #define DCC_ZERO_BIT_PULSE_DURATION_TIMER3 1599
  #define DCC_ONE_BIT_TOTAL_DURATION_TIMER3 1855
  #define DCC_ONE_BIT_PULSE_DURATION_TIMER3 927
  pinMode(DIRECTION_MOTOR_CHANNEL_PIN_B,INPUT);      // ensure this pin is not active! Direction will be controlled by DCC SIGNAL instead (below)
  digitalWrite(DIRECTION_MOTOR_CHANNEL_PIN_B,LOW);
  pinMode(DCC_SIGNAL_PIN_PROG,OUTPUT);      // THIS ARDUINO OUTPUT PIN MUST BE PHYSICALLY CONNECTED TO THE PIN FOR DIRECTION-B OF MOTOR CHANNEL-B
  bitSet(TCCR3A,WGM30);     // set Timer 3 to FAST PWM, with TOP=OCR3A
  bitSet(TCCR3A,WGM31);
  bitSet(TCCR3B,WGM32);
  bitSet(TCCR3B,WGM33);
  bitSet(TCCR3A,COM3B1);    // set Timer 3, OC3B (pin 2) to inverting toggle (actual direction is arbitrary)
  bitSet(TCCR3A,COM3B0);
  bitClear(TCCR3B,CS32);    // set Timer 3 prescale=1
  bitClear(TCCR3B,CS31);
  bitSet(TCCR3B,CS30);
  OCR3A=DCC_ONE_BIT_TOTAL_DURATION_TIMER3;
  OCR3B=DCC_ONE_BIT_PULSE_DURATION_TIMER3;
  pinMode(SIGNAL_ENABLE_PIN_PROG,OUTPUT);   // master enable for motor channel B
  progRegs.loadPacket(1,RegisterList::idlePacket,2,0);    // load idle packet into register 1    
  bitSet(TIMSK3,OCIE3B);    // enable interrupt vector for Timer 3 Output Compare B Match (OCR3B)    
#endif
} 

#define DCC_SIGNAL(R,N) \
  if(R.currentBit==R.currentReg->activePacket->nBits){    /* IF no more bits in this DCC Packet */\
      R.currentBit=0;                                     /*   reset current bit pointer and determine which Register and Packet to process next--- */\   
    if(R.nRepeat>0 && R.currentReg==R.reg){               /*   IF current Register is first Register AND should be repeated */\
      R.nRepeat--;                                        /*     decrement repeat count; result is this same Packet will be repeated */\
    } else if(R.nextReg!=NULL){                           /*   ELSE IF another Register has been updated */\
      R.currentReg=R.nextReg;                             /*     update currentReg to nextReg */\
      R.nextReg=NULL;                                     /*     reset nextReg to NULL */\
      R.tempPacket=R.currentReg->activePacket;            /*     flip active and update Packets */\        
      R.currentReg->activePacket=R.currentReg->updatePacket;\
      R.currentReg->updatePacket=R.tempPacket;\
    } else{                                               /*   ELSE simply move to next Register */\
      if(R.currentReg==R.maxLoadedReg)                    /*     BUT IF this is last Register loaded */\
        R.currentReg=R.reg;                               /*       first reset currentReg to base Register, THEN */\
        R.currentReg++;                                   /*     increment current Register (note this logic causes Register[0] to be skipped when simply cycling through all Registers) */\
    }                                                     /*   END-ELSE */\
  }                                                       /* END-IF: currentReg, activePacket, and currentBit should now be properly set to point to next DCC bit */\
  if(R.currentReg->activePacket->buf[R.currentBit/8] & R.bitMask[R.currentBit%8]){     /* IF bit is a ONE */ \
    OCR ## N ## A=DCC_ONE_BIT_TOTAL_DURATION_TIMER ## N;                               /*   set OCRA for timer N to full cycle duration of DCC ONE bit */\
    OCR ## N ## B=DCC_ONE_BIT_PULSE_DURATION_TIMER ## N;                               /*   set OCRB for timer N to half cycle duration of DCC ONE but */\
  } else{                                                                              /* ELSE it is a ZERO */\
    OCR ## N ## A=DCC_ZERO_BIT_TOTAL_DURATION_TIMER ## N;                              /*   set OCRA for timer N to full cycle duration of DCC ZERO bit */\
    OCR ## N ## B=DCC_ZERO_BIT_PULSE_DURATION_TIMER ## N;                              /*   set OCRB for timer N to half cycle duration of DCC ZERO bit */\
  }                                                                                    /* END-ELSE */\ 
    R.currentBit++;                                         /* point to next bit in current Packet */  
ISR(TIMER1_COMPB_vect){              // set interrupt service for OCR1B of TIMER-1 which flips direction bit of Motor Shield Channel A controlling Main Track
  DCC_SIGNAL(mainRegs,1)
}
#ifdef ARDUINO_AVR_UNO      // Configuration for UNO
ISR(TIMER0_COMPB_vect){              // set interrupt service for OCR1B of TIMER-0 which flips direction bit of Motor Shield Channel B controlling Prog Track
  DCC_SIGNAL(progRegs,0)
}
#else                       // Configuration para Arduino MEGA
ISR(TIMER3_COMPB_vect){              // set interrupt service for OCR3B of TIMER-3 which flips direction bit of Motor Shield Channel B controlling Prog Track
  DCC_SIGNAL(progRegs,3)
}
#endif

void showConfiguration(){
  int mac_address[]=MAC_ADDRESS;
  Serial.print("\n*** DCC++ Configuracsao ***\n");
  Serial.print("\nVersao:      ");
  Serial.print(VERSION);
  Serial.print("\nCompilado:     ");
  Serial.print(__DATE__);
  Serial.print(" ");
  Serial.print(__TIME__);
  Serial.print("\nArduino:      ");
  Serial.print(ARDUINO_TYPE);
  Serial.print("\n\nMotorShield: ");
  Serial.print(MOTOR_SHIELD_NAME);
  Serial.print("\n\nDCC Sig Main: ");
  Serial.print(DCC_SIGNAL_PIN_MAIN);
  Serial.print("\n   Direction: ");
  Serial.print(DIRECTION_MOTOR_CHANNEL_PIN_A);
  Serial.print("\n   Enable: ");
  Serial.print(SIGNAL_ENABLE_PIN_MAIN);
  Serial.print("\n   Monitor: ");
  Serial.print(CURRENT_MONITOR_PIN_MAIN);
  Serial.print("\n\nDCC Sig Prog: ");
  Serial.print(DCC_SIGNAL_PIN_PROG);
  Serial.print("\n   Direction: ");
  Serial.print(DIRECTION_MOTOR_CHANNEL_PIN_B);
  Serial.print("\n   Enable: ");
  Serial.print(SIGNAL_ENABLE_PIN_PROG);
  Serial.print("\n   Monitor: ");
  Serial.print(CURRENT_MONITOR_PIN_PROG);
  Serial.print("\n\nNum Turnouts: ");
  Serial.print(EEStore::eeStore->data.nTurnouts);
  Serial.print("\n   Sensores: ");
  Serial.print(EEStore::eeStore->data.nSensors);
  Serial.print("\n   Outputs: ");
  Serial.print(EEStore::eeStore->data.nOutputs);
  Serial.print("\n\nInterface: ");
  #if COMM_TYPE == 0
    Serial.print("Serial");
  #elif COMM_TYPE == 1
    Serial.print(COMM_SHIELD_NAME);
    Serial.print("\nMAC Address:  ");
    for(int i=0;i<5;i++){
    Serial.print(mac_address[i],HEX);
    Serial.print(":");
    }
    Serial.print(mac_address[5],HEX);
    Serial.print("\nPorta:         ");
    Serial.print(ETHERNET_PORT);
    Serial.print("\nIP Address:   ");
  #ifdef IP_ADDRESS
    Ethernet.begin(mac,IP_ADDRESS);           // Start networking using STATIC IP Address
  #else
    Ethernet.begin(mac);                      // Start networking using DHCP to get an IP Address
  #endif     
    Serial.print(Ethernet.localIP());
  #ifdef IP_ADDRESS
    Serial.print(" (Static)");
  #else
    Serial.print(" (DHCP)");
  #endif
#endif
    Serial.print("\n\nPrograma Parado - Resete o Arduino");
while(true);
}

// Joahermen - 18 de Abril de 2019
